package mmtext.images;

import java.awt.Image;
import java.awt.Point;

import mmtext.images.utils.ImageBuilder;

/**
 * Project: Computer Vision Framework
 * 
 * @author Wonder Alexandre Luz Alves
 * @advisor Ronaldo Fumio Hashimoto
 * 
 * @date 10/09/2007
 * 
 * @description Classe que representa uma imagem digital. Essa classe utiliza
 *              somente as APIs do java para manipular os pixels da imagens
 * 
 */
public class GrayScaleImage implements IGrayScaleImage {
	private int width; // largura da imagem
	private int height; // altura da imagem
	private int pixels[][]; // matriz de pixel da imagem

	public GrayScaleImage(int matrix[][]) {
		this.width = matrix.length;
		this.height = matrix[0].length;
		this.pixels = matrix;
	}

	/**
	 * Construtor para criar uma nova imagem
	 * 
	 * @param width
	 *            - largura
	 * @param height
	 *            - altura
	 */
	public GrayScaleImage(int width, int height) {
		this.width = width;
		this.height = height;
		this.pixels = new int[width][height];
	}

	public GrayScaleImage invert() {
		GrayScaleImage imgOut = new GrayScaleImage(this.getWidth(), this
				.getHeight());
		for (int w = 0; w < this.getWidth(); w++) {
			for (int h = 0; h < this.getHeight(); h++) {
				imgOut.setPixel(w, h, 255 - this.getPixel(w, h));
			}
		}
		return imgOut;
	}

	public int countPixels(int color) {
		int area = 0;
		for (int w = 0; w < getWidth(); w++) {
			for (int h = 0; h < getHeight(); h++) {
				if (getPixel(w, h) == color) {
					area++;
				}
			}
		}
		return area;
	}

	/**
	 * Inicializar todos os pixel da imagem para um dado nivel de cinza
	 * 
	 * @param color
	 */
	public void initImage(int color) {
		for (int i = 0; i < getWidth(); i++) {
			for (int j = 0; j < getHeight(); j++) {
				setPixel(i, j, color);
			}
		}
	}

	/**
	 * modifica todos os niveis de cinza de uma dada cor para uma outra cor
	 * 
	 * @param colorOld
	 * @param colorNew
	 */
	public void replaceColor(int colorOld, int colorNew) {
		for (int i = 0; i < this.getSize(); i++) {
			if (getPixel(i) == colorOld)
				setPixel(i, colorNew);
		}
	}

	/**
	 * Cria uma copia da imagem original
	 * 
	 * @return BinaryImage - nova imagem
	 */
	public GrayScaleImage duplicate() {
		GrayScaleImage clone = new GrayScaleImage(getWidth(), getHeight());
		for (int i = 0; i < getWidth(); i++) {
			for (int j = 0; j < getHeight(); j++) {
				clone.setPixel(i, j, getPixel(i, j));
			}
		}
		return clone;
	}

	/**
	 * Pega o valor do pixel (x, y)
	 * 
	 * @param x
	 *            - largura
	 * @param y
	 *            - altura
	 * @return float - valor do pixel
	 */
	public int getPixel(int x, int y) {
		if (x >= getWidth() || y >= getHeight())
			throw new RuntimeException("(x=" + x + ", y=" + y
					+ ") Nao existe essa posicao na matriz de pixels");
		return pixels[x][y];
	}

	/**
	 * Modifica o valor do pixel (x, y) = value
	 * 
	 * @param x
	 *            - largura
	 * @param y
	 *            - altura
	 * @param value
	 *            - valor do pixel
	 */
	public void setPixel(int x, int y, int value) {
		if (x >= getWidth() || y >= getHeight())
			throw new RuntimeException("(x=" + x + ", y=" + y
					+ ") Nao existe essa posicao na matriz de pixels");
		pixels[x][y] = value;
	}

	/**
	 * Pega uma matriz bidimensional de pixel da imagem
	 * 
	 * @return int[][]
	 */
	public int[][] getPixels() {
		return pixels;
	}

	/**
	 * Modifica a matriz de pixel da imagem para os valores da matriz dada
	 * 
	 * @param matrix
	 */
	public void setPixels(int matrix[][]) {
		this.width = matrix.length;
		this.height = matrix[0].length;
		this.pixels = new int[width][height];
		for (int i = 0; i < getWidth(); i++) {
			for (int j = 0; j < getHeight(); j++) {
				setPixel(i, j, matrix[i][j]);
			}
		}
	}

	/**
	 * Modificao tamanho da matriz
	 */
	public void resize(int width, int height) {
		this.width = width;
		this.height = height;
		int matrix[][] = getPixels();
		this.pixels = new int[width][height];
		initImage(255);
		for (int i = 0; i < matrix.length; i++) {
			for (int j = 0; j < matrix[i].length; j++) {
				setPixel(i, j, matrix[i][j]);
			}
		}
	}

	/**
	 * Modificao tamanho da matriz
	 */
	public void resizeCenter(int width, int height) {
		this.width = width;
		this.height = height;
		int matrix[][] = getPixels();
		this.pixels = new int[width][height];
		initImage(255);

		for (int i = 0, x = Math.abs(matrix.length - width) / 2; i < matrix.length; i++, x++) {
			for (int j = 0, y = Math.abs(matrix.length - height) / 2; j < matrix[i].length; j++, y++) {
				setPixel(x, y, matrix[i][j]);
			}
		}
	}

	/**
	 * Retorna a altura da imagem
	 * 
	 * @return int - altura
	 */
	public int getHeight() {
		return height;
	}

	/**
	 * Retorna a largura da imagem
	 * 
	 * @return int - largura
	 */
	public int getWidth() {
		return width;
	}

	/**
	 * Retorna o tamanho da imagem
	 * 
	 * @return
	 */
	public int getSize() {
		return getHeight() * getWidth();
	}

	/**
	 * Converte uma imagem de gray scale para binaria
	 * 
	 * @param limiar
	 *            - limiar do threshold
	 * @param background
	 *            - background da imagem binaria
	 */
	public IBinaryImage convertBinaryImage(int limiar, int backgound) {
		IBinaryImage imgB = new BinaryImage(getWidth(), getHeight());
		imgB.setBackground(backgound);
		for (int i = 0; i < this.getWidth(); i++) {
			for (int j = 0; j < this.getHeight(); j++) {
				if (this.getPixel(i, j) > limiar)
					imgB.setPixel(i, j, imgB.getBackground());
				else
					imgB.setPixel(i, j, imgB.getForeground());
			}
		}
		return imgB;
	}

	/**
	 * Pega o maior pixel da imagem
	 */
	public int getPixelMax() {
		int max = Integer.MIN_VALUE;
		for (int i = 0; i < getWidth(); i++) {
			for (int j = 0; j < getHeight(); j++) {
				if (getPixel(i, j) > max)
					max = getPixel(i, j);
			}
		}
		return max;
	}

	/**
	 * Pega o menor pixel da imagem
	 */
	public int getPixelMin() {
		int min = Integer.MAX_VALUE;
		for (int i = 0; i < getWidth(); i++) {
			for (int j = 0; j < getHeight(); j++) {
				if (getPixel(i, j) < min)
					min = getPixel(i, j);
			}
		}

		return min;
	}

	/**
	 * Pega um histograma da imagem
	 * 
	 * @return int[]
	 */
	public int[] getHistogram() {
		int result[] = new int[256];
		for (int w = 0; w < getWidth(); w++) {
			for (int h = 0; h < getHeight(); h++) {
				result[getPixel(w, h)]++;
			}
		}
		return result;
	}

	/**
	 * Pega o numero de pixel diferente na imagem
	 */
	public int numPixels() {
		int h[] = getHistogram();
		int numPixel = 0;
		for (int i = 0; i < h.length; i++)
			if (h[i] > 0)
				numPixel++;
		return numPixel;

	}

	/**
	 * Verifica se duas imagens sao iguais
	 * 
	 * @param img
	 *            - IGrayScaleImage
	 * @return true se forem iguais false caso contrario
	 */
	public boolean equals(Object o) {
		GrayScaleImage img = (GrayScaleImage) o;
		for (int x = 0; x < getWidth(); x++)
			for (int y = 0; y < getHeight(); y++)
				if (getPixel(x, y) != img.getPixel(x, y))
					return false;

		return true;
	}

	public void add(int a) {
		for (int x = 0; x < getWidth(); x++) {
			for (int y = 0; y < getHeight(); y++) {
				setPixel(x, y, getPixel(x, y) + a);
			}
		}

	}

	public void multiply(int a) {
		for (int x = 0; x < getWidth(); x++) {
			for (int y = 0; y < getHeight(); y++) {
				setPixel(x, y, getPixel(x, y) * a);
			}
		}

	}

	public void setPixel(int i, int level) {
		int w = i / getHeight();
		int h = i % getHeight();
		pixels[w][h] = level;
	}

	public int getPixel(int i) {
		int w = i / getHeight();
		int h = i % getHeight();
		return pixels[w][h];
	}

	public Image createImage() {
		return ImageBuilder.convertToImage(pixels);
	}

	/**
	 * Pegando um boundbox
	 * 
	 * @param x1
	 * @param y1
	 * @param x2
	 * @param y2
	 * @return
	 */
	public BoundBox getBoundBox(int x1, int y1, int x2, int y2, int fore) {
		BoundBox out = new BoundBox() {
			private int x1, y1, x2, y2;
			public int fore;

			public int getPixelForeground() {
				return fore;
			}

			public void setPixelForeground(int c) {
				fore = c;
			}

			public Point getPointBegin() {
				return new Point(x1, y1);
			}

			public Point getPointEnd() {
				return new Point(x2, y2);
			}

			public void setDimension(int x1, int y1, int x2, int y2) {
				// if(x1 >= getWidth() || y1 >= getHeight() || x2 >= getWidth()
				// || y2 >= getHeight()) throw new
				// RuntimeException("Tamanho do boundbox invalido!");
				this.x1 = x1;
				this.x2 = x2;
				this.y1 = y1;
				this.y2 = y2;
			}

			public void setDimension(int x1, int y1, int x2, int y2, int fore) {
				// if(x1 >= getWidth() || y1 >= getHeight() || x2 >= getWidth()
				// || y2 >= getHeight()) throw new
				// RuntimeException("Tamanho do boundbox invalido!");
				this.x1 = x1;
				this.x2 = x2;
				this.y1 = y1;
				this.y2 = y2;
				this.fore = fore;
			}

			public int getPixel(int w, int h) {
				w += x1;
				h += y1;
				if (fore == pixels[w][h] || fore == -1)
					return pixels[w][h];
				return -1;
			}

			public void setPixel(int w, int h, int value) {
				// if(w < x1 || w > x2 || h < y1 || h > y2) throw new
				// RuntimeException("(x="+w+", y="+h+") Nao existe essa posicao no boundbox");
				w += x1;
				h += y1;
				pixels[w][h] = value;
			}

			public void paintBoundBox(int color) {
				int w = x2 - x1;
				int h = y2 - y1;
				for (int i = 0; i < w; i++) {
					for (int j = 0; j < h; j++) {
						if (i == 0 || j == 0 || i == w - 1 || j == h - 1)
							pixels[x1 + i][y1 + j] = color;
					}
				}
			}

			public IGrayScaleImage getImage() {
				IGrayScaleImage out = new GrayScaleImage(Math.abs(x2 - x1),
						Math.abs(y2 - y1));
				for (int i = 0; i < out.getWidth(); i++) {
					for (int j = 0; j < out.getHeight(); j++) {
						out.setPixel(i, j, getPixel(i, j));
					}
				}
				return out;
			}

			public int getWidth() {
				return x2 - x1 + 1;
			}

			public int getHeight() {
				return y2 - y1 + 1;
			}

			public int countPixels(int color) {
				int area = 0;
				for (int w = 0; w < getWidth(); w++) {
					for (int h = 0; h < getHeight(); h++) {
						if (getPixel(w, h) == color) {
							area++;
						}
					}
				}
				return area;
			}

			public void initPixels(int color) {
				for (int i = 0; i < getWidth(); i++) {
					for (int j = 0; j < getHeight(); j++)
						setPixel(i, j, color);
				}
			}

		};
		out.setDimension(x1, y1, x2, y2, fore);

		return out;
	}

}
